"""
This is one of 2 files required to run the neural chatbot in spinnaker.  The other is SimpleChatbotAgentSpinn8.py.
This file contains the interface (world). SimpleChatbotAgentSpinn8.py contains the chatbot (agent).

To run the chatbot you just need to run SimpleChatbotAgentSpinn8.py as it will call the relevant functions in this file
(ChatBotInterfaceSpinn8.py).

This file contains functions used to communicate with the agent (i.e. the spinnaker board) via live injectors and receivers.

When you run SimpleChatbotAgentSpinn8.py, an "enter a question" prompt should appear asking you for a query.  
  
Currently, you can enter one of 3 queries:
Where is the yellow banana?
Where is the red apple?
What shape is the diamond?

The language, i.e. the sentences the agent can use to answer the queries, are defined in chatbotLanguage.txt
The answers to the above queries are programmed/defined in LanguageGenerator.py.


A query entered at the interface is sent
from the environment (interface) to the spinnaker board (agent/chatbot) via
a spike which activates a query cell.  The agent (spinnaker board) then processes the query and sends back an answer (sentence)
to the environment (interface) via another set of spikes.  These spikes are fired by neurons representing words, where each neuron represents a different word
(the words represented are shown in OutputWords.txt). Therefore, the interface builds the response (sentence) using the words in OutputWords.txt and
the spikes, and spike times, from the word neurons.

Communication is achieved via live injection (from world to agent) and
receivers (from agent to world) of spikes.

"""

import time
import numpy as np
import os.path
import spynnaker8.external_devices as externaldevices



Words =[] #stores the words from file "OutputWords.txt".  These are all of the different words comprising the
          #language (sentences) in chatbotLanguage.txt
wordNeuronsFired=[]  ##stores the spike times of the word neurons that fired
wordSpikeTimes=[]    ##stores the word neurons that fired

AgentReady=1          #tells us the system state, 0 if agent is processing response,
                      #1 if system is ready for next query, and 2 if agent has quit

Generator=-1   #contains a generator object passed from SimpleChatbotAgentSpinn8.py
    

def GetAndSendQuery(spikesToSpinnConn):
    #this sends the query from the user to the spinnaker board, i.e. causes the query cell associated to the given query to
    #fire via live injection.
    global AgentReady
    validquestion=0
    
    while(validquestion==0):
        
        query=raw_input("Enter a question: ")
        time.sleep(1) ##wait an aribitrary amount of time to determine whether the agent has finished
        if(AgentReady!=2): ##if agent is still running
            if(query=="where is the yellow banana?"):
                sendQuery=0 # query neuron 0 corresponds to "where is the yellow banana?"
                spikesToSpinnConn.send_spike("QueryToBoard",
                    sendQuery,send_full_keys=True)
                validquestion=1
                AgentReady=0
                print "sent Query, waiting for response...."
            elif(query=="where is the red apple?"):
                sendQuery=1 # query neuron 1 corresponds to "where is the red apple?"
                print "sent Query, waiting for response...."
                spikesToSpinnConn.send_spike("QueryToBoard",
                    sendQuery,send_full_keys=True)
                validquestion=1
                AgentReady=0
            elif(query=="what shape is the diamond?"):
                sendQuery=2 # query neuron 2 corresponds to "what shape is the diamond?"
                spikesToSpinnConn.send_spike("QueryToBoard",
                    sendQuery,send_full_keys=True)
                validquestion=1
                AgentReady=0
                print "sent Query, waiting for response...."
            else:
                print "not a valid Question, try again"
        else:
            break   #agent has stopped

def getAgentResponse(label, time, neuron_ids):
    ##this listens for the agent's response to the current query, i.e. listens for spikes from the word
    ##neurons on the spinnaker board, and, records the neurons that fired and the spike times.  When the interface has received the full response
    #from the agent, i.e. the neuron associated to a full stop has fired, the function prints out the sentence.

    global wordSpikeTimes
    global wordNeuronsFired
    global Words
    global AgentReady
    global Generator

    for neuron_id in neuron_ids:
        wordSpikeTimes+=[time]
        wordNeuronsFired+=[neuron_id]
        if(Words[neuron_id]=="."): #if the interface has received the full response from the agent, i.e. the word neuron associated to a full stop has fired,
                        #print out the sentence received and clear the stored spike times and neurons that fired ready for the next answer.
            Generator.printSentenceLive(wordNeuronsFired,wordSpikeTimes) #print the sentence
            
            wordSpikeTimes=[]
            wordNeuronsFired=[]
            AgentReady=1  #agent has sent full response


def getAndSendQueries(label,sender):
    #this allows a user to enter a query, which is sent via live injection to the spinnaker board. When the interface receives a
    #full response (notified by AgentReady) it asks the user for another question.

    global AgentReady
    timeScaleFactor=10  #set to the same value as the one in .spynnaker.cfg
    
    time.sleep(0.1*timeScaleFactor) # wait for an arbitatry amount of time for the system to set up
 
    while True: 
    
        if(AgentReady==1):   #if we have received a full response from the agent

            GetAndSendQuery(sender) #get and send the next query

        elif(AgentReady==2):  #if agent has finished
            break

        time.sleep(0.5*timeScaleFactor) # wait for an arbitatry amount of time, i.e. wait for the agent's responses

def QuitInterface(label, sender):
    
    global AgentReady
    AgentReady=2 #this is to notify the interface the agent has quit (i.e. the spinnaker simulation has ended).  

def RunInterface(generatorWordCells, generator):
    #this is called by SimpleChatbotAgentSpinn8.py.
    #this sets up communication between the world and spinnaker (live injectiors and receivers)
    #queries are injected into the spinnaker board via live injections
    #the interface listens for spikes from generatorWordCells, i.e. the word neurons
    global Words
    global Generator
    Generator=generator
    Words=Generator.RetrieveWords()   ##Read in all of the different words used by the agent


    externaldevices.activate_live_output_for(generatorWordCells,
                    database_notify_port_num=19996)             #set up listening

    spikesToSpinnConn = externaldevices.SpynnakerLiveSpikesConnection(
                        local_port=19999,
                        send_labels=["QueryToBoard"])           #set up injecting
                        



    
    spikesToSpinnRec = externaldevices.SpynnakerLiveSpikesConnection(               #set up listening
                        receive_labels=[generatorWordCells.label],local_port=19996)

    spikesToSpinnRec.add_receive_callback(generatorWordCells.label,getAgentResponse) #call getAgentResponse when spikes from the word population
                                                                                    #(generatorWordCells) are received
                                                                                    
                                                                                        

    spikesToSpinnConn.add_start_resume_callback(
                        "QueryToBoard", getAndSendQueries)  #send queries using function getAndSendQueries to the Query population labelled QueryToBoard

    spikesToSpinnConn.add_pause_stop_callback("QueryToBoard", QuitInterface) # when the simulation (agent) finishes, exit interface





